import functools
import re
import logging
import uuid

from typing import List, Callable, Union, Dict, Optional

from ppa import ui

from . import groups, exceptions, _object_classes
from ._object_classes import User, Group, Computer

logger = logging.getLogger(__name__)


def _with_retry(retries: int) -> Callable:
    logger.info(f"decorating function with {retries} retries")

    def _execute_with_retry(func: Callable) -> Callable:
        @functools.wraps(func)
        def _execute(*args, **kwargs):
            for i in range(0, retries):
                try:
                    result = func(*args, **kwargs)
                    assert result
                    return result
                except AssertionError:
                    if i != retries - 1:
                        ui.output_error("No results found, please try again.")

            message = f"Failed to find any records."
            ui.output_error(message)
            raise exceptions.NoSuchObject(message)

        return _execute

    return _execute_with_retry


def _match(*, pattern: str, value: str):
    try:
        return re.match(pattern, value)
    except re.error:
        raise exceptions.ParameterError(f"Supplied filter '{pattern}' is not a valid regular expression.")


def _filter_record(record: Union[User, Group, Computer], criteria: Dict[str, str], filter_func: Callable) -> bool:
    """
    - Look through the attributes and patterns in criteria.
    - Attempt to match against each pattern against the record attribute.
    - Return False if the record produces a match.
    """
    if criteria:
        for attribute, pattern in criteria.items():
            try:
                record_value = getattr(record, attribute)
            except AttributeError:
                message = f"Attempted to filter by invalid or missing LDAP attribute '{attribute}'."
                ui.output_error(message)
                raise exceptions.ParameterError(message)

            try:
                match = bool(filter_func(pattern=pattern, value=record_value))
                if match:
                    logger.debug(
                        f"filtering record as '{attribute}' value " f"'{record_value}' matches pattern '{pattern}'"
                    )
                    return False
            except exceptions.ParameterError as e:
                ui.output_error(str(e))
                raise e
    return True


def _exclude(records: list, criteria: Union[dict, None]) -> list:
    """
    - Pass each record through the supplied filter function, excluding
        records that produce a match.
    - Return the filtered results.
    """
    return list(filter(lambda x: _filter_record(x, criteria, _match), records))


def _choose_from_table(*, header: list, data: List[list], min_records: int, max_records: int) -> list:
    """
    - Presents a table using the supplied header and data.
    - Runs until the user chooses a valid number of entries.
    - Returns the selected indices.
    """
    hint: Union[int, str]
    table_id = str(uuid.uuid4())
    if min_records > max_records:
        # Indicates no upper limit, number of chosen
        # records just needs to be more than the minimum.
        hint = f"{min_records if min_records > 0 else 1} or more"
        validator = functools.partial(lambda x: x >= min_records)
    else:
        if all([min_records == max_records, min_records > 0]):
            hint = min_records
            validator = functools.partial(lambda x: x == min_records)
        elif all([min_records == 0, max_records == 1]):
            # get_user and get_group should have no more than
            # one record selected by the user, but the hint
            # text should be "select one records(s)" because
            # it shows the correct intent of the function. The
            # user needs to be able to choose 0 so they can
            # retry with a different value.
            # https://github.com/Osirium/opus-tasks/issues/71
            hint = "1"
            validator = functools.partial(lambda x: x in {0, 1})
        else:
            hint = f"{min_records}-{max_records}"
            # Number of chosen records needs to be in min-max range.
            validator = functools.partial(lambda x: x in range(min_records, max_records + 1))
    while True:
        indices = ui.input_table(f"Select {hint} Record(s)", header=header, data=data, element_id=table_id)
        if not validator(len(indices)):
            ui.output_error(f"Please Select {hint} Record(s).")
        else:
            break
    return indices


def _pick_records(*, records: list, min_records: int, max_records: Union[int, None], auto_select: bool) -> list:
    """
    - Takes a master record list, a table header, and table data.
    - Calls _choose_from_table to get the record indices chosen by the user.
    - Returns a subset of the master records based on the chosen indices.
    """
    if len(records) == 1:
        if auto_select:
            logger.debug("auto-selecting single record from search results")
            ui.output_info(
                f"Selected {records[0]._ppa_record_name} is "  # type: ignore
                f"{getattr(records[0], 'cn') or getattr(records[0], 'sAMAccountName')}"
            )
            return records

        else:
            logger.debug("prompting for confirmation before returning single record from search results")
            table_type = records[0]._ppa_table_type  # type: ignore
            record_type = records[0]._ppa_record_name.title()  # type: ignore
            ui.output_table(
                f"1 {record_type} Record Found",
                header=table_type.header,
                data=[
                    [
                        # If an attribute is None it should be empty in the table, otherwise show
                        # its string value (for example some values are False, and these should be displayed).
                        str(getattr(records[0], col)) if getattr(records[0], col) is not None else ""
                        for col in table_type.columns
                    ]
                ],
            )
            if ui.input_confirm(f"Select {record_type} Record?", confirm="Yes", cancel="No"):
                return records
            return []

    else:
        record_types = list(set(type(r) for r in records))  # type: ignore
        if len(record_types) != 1:
            raise exceptions.ParameterError(f"Tried to generate a table containing {len(record_types)} record types.")
        table_type = records[0]._ppa_table_type
        return [
            records[i]
            for i in _choose_from_table(
                header=table_type.header,
                data=[
                    # If an attribute is None it should be empty in the table, otherwise show
                    # its string value (for example some values are False, and these should be displayed).
                    [str(getattr(rec, col)) if getattr(rec, col) is not None else "" for col in table_type.columns]
                    for rec in records
                ],
                min_records=min_records,
                max_records=max_records if max_records else -1,  # default to no upper limit
            )
        ]


def _get_object(
    connection,
    class_display_name: str,
    attr_dicts: List[Dict[str, str]],
    search_function: Callable,
    search_attribute: Optional[str] = None,
    search_base: Optional[str] = None,
    exclude: Optional[Dict[str, str]] = None,
    auto_select: Optional[bool] = False,
    auto_wildcard: Optional[bool] = False,
    title: Optional[str] = None,
):
    class_display_name_lower = class_display_name.lower()

    header = f"### {title}" if title else f"### Get Active Directory {class_display_name}"
    ui.output_markdown(header)

    if not search_attribute:
        choice = ui.input_choice(
            f"({class_display_name} Search) - Select a search attribute",
            options=sorted([d["display"] for d in attr_dicts]),
        )
        search_attribute = next((a["value"] for a in attr_dicts if a["display"] == choice))

    # Try to establish a nice display name for the attribute from the attr_dicts
    # parameter, but if we can't, just fall back to the real attribute name.
    try:
        search_attribute_display = [d["display"] for d in attr_dicts if d["value"] == search_attribute][0]
    except (KeyError, IndexError):
        search_attribute_display = search_attribute

    @_with_retry(retries=10)
    def _search():
        def _get_value():
            value = ui.input_text(
                f"({class_display_name} Search) - {search_attribute_display}",
                required=True,
                element_id=search_attribute,
            )
            if auto_wildcard:
                value = f"{value}*" if not value.endswith("*") else value
            return value

        search_results = search_function(connection, search_base=search_base, **{search_attribute: _get_value()})
        return _exclude(records=search_results, criteria=exclude)

    while True:
        try:
            filtered_record = _pick_records(records=_search(), min_records=0, max_records=1, auto_select=auto_select)
        except exceptions.NoSuchObject as e:
            raise e

        if not filtered_record:
            # The user selected no search results, let them search again, otherwise they have
            # to call the function over and over inside the task.
            ui.output_info(
                f"No {class_display_name_lower} has been chosen from this search, please try again",
                element_id=f"no_{class_display_name_lower}_chosen",
            )
        else:
            # A result was chosen.
            break

    return next(iter(filtered_record))


def _add_object_to_groups(
    connection,
    object_module,
    object_type_str,
    *,
    distinguished_name: str,
    search_attribute: Optional[str] = None,
    exclude: Optional[Dict[str, str]] = None,
    auto_select: Optional[bool] = False,
) -> tuple:
    capitalized_type_str = object_type_str.capitalize()
    attr_dicts = [{"display": "SAM Account Name", "value": "sAMAccountName"}, {"display": "Common Name", "value": "cn"}]
    ui.output_markdown(
        f"""### Add Active Directory {capitalized_type_str} to Security Groups

__Operating on {capitalized_type_str}__: *{distinguished_name}*
"""
    )

    if not object_module.exists(connection, distinguishedName=distinguished_name):
        message = f"{capitalized_type_str} with distinguished name {distinguished_name} does not exist."
        ui.output_error(message)
        raise exceptions.NoSuchObject(message)

    if not search_attribute:
        choice = ui.input_choice(
            f"(Group Search) - Select a search attribute", options=sorted([d["display"] for d in attr_dicts])
        )
        search_attribute = next((a["value"] for a in attr_dicts if a["display"] == choice))

    # Try to establish a nice display name for the attribute from the attr_dicts
    # parameter, but if we can't, just fall back to the real attribute name.
    try:
        search_attribute_display = [d["display"] for d in attr_dicts if d["value"] == search_attribute][0]
    except (KeyError, IndexError):
        search_attribute_display = search_attribute

    @_with_retry(retries=10)
    def _search() -> List[Group]:
        search_params = {
            search_attribute: ui.input_text(
                f"(Group Search) - {search_attribute_display}", required=True, element_id=search_attribute
            )
        }
        search_results = groups.search(connection, **search_params)
        return _exclude(records=search_results, criteria=exclude)

    def _select_groups() -> List[Group]:
        while True:
            search_results = _search()
            chosen_records = _pick_records(
                records=search_results, min_records=0, max_records=-1, auto_select=auto_select
            )
            if not chosen_records:
                ui.output_error(
                    "No groups have been chosen from this search. Please try again.", element_id="no_groups_chosen"
                )
            else:
                break
        return chosen_records

    chosen_groups: List[Group] = []
    while True:
        if len(chosen_groups) >= 1:
            ui.output_info(
                f"Total Groups Selected: {len(chosen_groups)}", element_id=f"{distinguished_name}_selected_groups"
            )
        selection = _select_groups()
        if any([len(selection) > 1, all([len(selection) == 1, auto_select])]):
            # - If there is only one group and auto_select is False, the user
            #   will have confirmed the single record already.
            #
            # - If there is more than one group, we should pop up a confirm table.
            #
            # - If there is only one group and auto_select is True, we should
            #   pop up a confirm table.
            ui.output_table(
                "Selected Groups",
                header=_object_classes._group_table.header,
                data=[
                    [getattr(group, col) or "" for col in _object_classes._group_table.columns] for group in selection
                ],
                element_id=f"{distinguished_name}_confirm_selected_groups",
            )
            if not ui.input_confirm(
                "Confirm or Search Again?",
                confirm="Confirm",
                cancel="Search Again",
                element_id=f"{distinguished_name}_confirm_group",
            ):
                # Start the search again.
                continue

        chosen_groups.extend([s for s in selection if s not in chosen_groups])
        if ui.input_confirm(
            "(Confirm Groups) - Continue Selecting, or Proceed?",
            confirm="Proceed",
            cancel="Continue Selecting",
            element_id=f"{distinguished_name}_more_or_proceed",
        ):
            break

    display = "\n".join([f"* {getattr(group, 'name') or getattr(group, 'sAMAccountName')}" for group in chosen_groups])

    ui.output_markdown(
        f"""#### Click 'Continue' to add the {object_type_str} to the following groups:

{display}
"""
    )
    ui.input_accept("Continue")

    existing_memberships = []
    ui.output_progress(
        f"(Perform Operation) - Adding {object_type_str} to {len(chosen_groups)} groups...",
        percent=None,
        element_id=f"{distinguished_name}_add_operation",
    )
    total = len(chosen_groups)
    increment = int(100 / total)
    progress = 0
    for i, group in enumerate(chosen_groups, 1):
        ui.output_progress(
            f"(Perform Operation) - Adding {object_type_str} to group {i} of {total}...",
            percent=progress,
            element_id=f"{distinguished_name}_add_operation",
        )
        try:
            groups.add_member(
                connection,
                distinguished_name=group.distinguishedName,  # type: ignore
                member_distinguished_name=distinguished_name,
            )
        except exceptions.ObjectExists:
            ui.output_error(
                f"{capitalized_type_str} is already a member of group "
                f"{getattr(group, 'name') or getattr(group, 'sAMAccountName')} - ignoring."
            )
            existing_memberships.append(group)
        progress += increment

    ui.output_progress(
        f"(Perform Operation) - {capitalized_type_str} has been added to the {total} chosen group(s).",
        percent=100,
        element_id=f"{distinguished_name}_add_operation",
    )
    ui.output_markdown("Operation completed successfully")
    ui.input_accept("Ok")

    groups_added_to = [group for group in chosen_groups if group not in existing_memberships]
    return (groups_added_to, existing_memberships)


def _remove_object_from_groups(
    connection,
    object_module,
    object_type_str,
    *,
    distinguished_name: str,
    exclude: Optional[Dict[str, str]] = None,
    auto_select: Optional[bool] = False,
) -> List[Group]:
    capitalized_type_str = object_type_str.capitalize()
    ui.output_markdown(
        f"""### Remove Active Directory {capitalized_type_str} from Security Groups

__Operating on {capitalized_type_str}__: name *{distinguished_name}*
    """
    )

    def _search():
        try:
            filtered_groups = _exclude(
                records=object_module.group_memberships(connection=connection, distinguished_name=distinguished_name),
                criteria=exclude,
            )
        except exceptions.NoSuchObject as e:
            ui.output_error(f"{capitalized_type_str} with distinguished name {distinguished_name} does not exist.")
            raise e
        if not filtered_groups:
            message = (
                f"Either the {object_type_str} is not in any groups, or those groups have been excluded from this task."
            )
            ui.output_error(message)
            raise exceptions.OperationFailed(message)
        return filtered_groups

    filtered_groups = _search()
    chosen_groups: List[Group] = _pick_records(
        records=filtered_groups, min_records=1, max_records=-1, auto_select=auto_select
    )

    if not chosen_groups:
        ui.output_error("Operation cancelled, no groups have been selected.")
        return []

    display = "\n".join([f"* {getattr(group, 'name') or getattr(group, 'sAMAccountName')}" for group in chosen_groups])

    ui.output_markdown(
        f"""#### Click 'Continue' to remove the {object_type_str} from the following groups:

{display}
"""
    )
    ui.input_accept("Continue")

    total = len(chosen_groups)
    progress = 0
    increment = int(100 / total)
    for i, group in enumerate(chosen_groups, 1):
        ui.output_progress(
            f"Removing {object_type_str} from group {i} of {total}...",
            percent=progress,
            element_id=f"{distinguished_name}_remove_operation",
        )
        groups.remove_member(
            connection,
            distinguished_name=group.distinguishedName,  # type: ignore
            member_distinguished_name=distinguished_name,
        )
        progress += increment
    ui.output_progress(
        f"{capitalized_type_str} has been removed from the {total} chosen group(s).",
        percent=100,
        element_id=f"{distinguished_name}_remove_operation",
    )
    ui.output_markdown(f"Operation completed successfully")
    ui.input_accept("Ok")
    return chosen_groups
